/*
 sbcp_math.c
  math routines required for sbcp (version 0.1.6)

  Created: November 29, 2007 4:59 PM EDST
  Revised: December 22, 2007 10:51 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Saturday, December 8, 2007 11:20 PM
A.P.: moved sbcp_parse_number here (to reduce compilation time in getting it to work).
Wednesday, December 19, 2007 6:41 PM
  this source now avoids the use of FPU operations, altogether.
Friday, December 21, 2007 12:57 PM
  _int64_neg is now in the main header as a macro (because it is simple enough).
*/

#include <fp.h>

#if TYPE_LONGLONG
# undef TYPE_LONGLONG
# define TYPE_LONGLONG			0
#endif

#define _int64_param			struct _int64*
#define _uint64_param			struct _uint64*

#include "sbcp.h"
#include "sbcp_utils.h"

#ifdef env68k
# pragma segment sbcp
#endif

/* ---------- glue code */

#if !defined(env68k) /* handled by sbcp_execution.a */
long int _int64_to_l(_int64 *n)
{
	if (int64hi(*n)>=0)
	{
		if (int64hi(*n)!=0 || (int64lo(*n)&0x80000000))
			return (long)2147483647;
	} else if (int64hi(*n)<-1 || !(int64lo(*n)&0x80000000))
		return (long)-2147483648;

	return int64lo(*n);
}

# if envppc /* parameters are registers */
#  define ah					(a>>16)
#  define bh					(b>>16)
#  define al					(_uint16)a
#  define bl					(_uint16)b
# else
#  define ah					(0+((_uint16*)&a)[_ERM])
#  define bh					(0+((_uint16*)&b)[_ERM])
#  define al					(0+((_uint16*)&a)[1-_ERM])
#  define bl					(0+((_uint16*)&b)[1-_ERM])
# endif

void _mul64_s32(_int64 *n, _int32 a, _int32 b)
{
	_uint32			m = (signed)ah*bl+al*(signed)bh, h = (signed)ah*(signed)bh;
	_uint32			l = al*bl, t = l;

	if ((l+=m<<16)<t)
		++h;
	int64_sethl(*n,h+=(signed)m>>16,l);
}

void _mul64_u32(_uint64 *n, _uint32 a, _uint32 b)
{
	_uint32			m = ah*bl+al*bh, h = ah*bh, l = al*bl, t = l;

	if ((l+=m<<16)<t)
		++h;
	int64_sethl(*n,h+=m>>16,l);
}

# undef bh
# undef bl
# if envppc /* parameters are registers */
#  define bh					(int64lo(n)>>16)
#  define bl					(_uint16)int64lo(n)
# else
#  define bh					(0+((_uint16*)&n->lo)[_ERM])
#  define bl					(0+((_uint16*)&n->lo)[1-_ERM])
# endif

void _int64_mulu32(_uint64 *n, _uint32 a)
{
	_uint32			m = ah*bl+al*bh, h = ah*bh+int64hi(n)*a, l = al*bl, t = l;

	if ((l+=m<<16)<t)
		++h;
	int64_sethl(*n,h+=m>>16,l);
}

# undef ah
# undef bh
# undef al
# undef bl
#endif /*!defined(env68k) AND !TYPE_LONGLONG*/

#if !defined(env68k) || !defined(envfpu)
void _int64_setx(_int64 *n, long double x)
{
	long double		y;

	if (x!=x || x<=-9223372036854775808.0) /* NAN or <=SINT64MIN */
		{int64_sethl(*n,-2147483648L,0U); return;}
	if (x>=9223372036854775807.0) /* >=SINT64MAX */
		{int64_sethl(*n,2147483647L,-1U); return;}
	y = x*(1.0/4294967296.0);
	int64_sethl(*n,y,x-y*4294967296.0);
}

void _uint64_setx(_uint64 *n, long double x)
{
	long double		y;

	if (x!=x || x>=18446744073709551615.0) /* NAN or >=UINT64MAX */
		{int64_sethl(*n,-1L,-1U); return;}
	if (x<=0) /* should it be (x<=-1)? */
		{int64_sethl(*n,0,0); return;}
	y = x*(1.0/4294967296.0);
	int64_sethl(*n,y,x-y*4294967296.0);
}
#endif

/* ---------- sbcp math code */

#ifndef env68k
/* Warning: this only checks for overflow; underflow is still okayed. */
boolean sbcp_float24_in_range_check(double v)
{
	/*return v==v && fabs(v)<=(double)3.689348815E+19;*/
	return (((_uint8*)&v)[_ERM*7]&0x7F)<=(_uint8)0x43;
}

_uint32 sbcp_make_float24(double v)
{ /* no more branches! */
	_int32			m = ((_int32*)&v)[_ERM], n = (0x80000000L&m)>>8;

	/* check the exponent range before rounding; take INF/NAN as out-of-range */
	m &= 0x7FFFFFFFL;
	m &= -((m-=0x43FFFFFFL)<0), m += 0x43FFFFFFL; /* min(x,0x43FFFFFFL) */

	m = m+(8-0x3C000000)>>4; /* matching data is 0Xxxxmmmmf */

	m &= -((m+=-0x7FFFFFL)<0), m += 0x7FFFFFL; /* min(m,0x7FFFFFL) */
	m &= -(m>=0); /* max(m,0) */

	return m|n;
}

void float24_to_double(_uint32 x, double *v)
{ /* no more usage of scalb */
	_int32			m = x, n = (x<<8)&0x80000000L, k = -((m=(m<<4)&0x7FFFFF00UL)!=0);

	k &= (m+=0x3C000000L), k |= n;
	((_int32*)v)[1-_ERM] = 0;
	((_int32*)v)[_ERM] = k;
}
#endif

/* ---------- number processing */

#if !defined(env68k) || !defined(envfpu)
const char *sbcp_parse_number(const char *s,
	const char *e,
	double *_v,
	boolean r)
{
	_uint8			c, n = 0;
	double			v = 0;
	int				u = 0, a = 0, b = 10, m = 0, x = 0;

	/* skip whitespace */
	do
	{
		if (s>=e)
			return NULL;
#if DBLSP_CHAR!=' '
	} while((c=*s++)<=' ' || c==DBLSP_CHAR);
#else
	} while((c=*s++)<=' ');
#endif

#if MINUS_CHAR!='-'
	if (c=='-' || c==(_uint8)MINUS_CHAR)
#else
	if (c=='-')
#endif
	{
		n = 1;
		if (s>=e)
			return NULL;
		c = *s++;
	} else if (c=='+')
	{
		if (s>=e)
			return NULL;
		c = *s++;
	} else if (c!='.' && (_uint8)(c-'0')>(_uint8)('9'-'0'))
		return NULL; /* will not process symbol ids or special characters here */

	if (c=='0' && s<e)
	{
		if ((c=*s-'X')==0 || (c+='X'-'x')==0)
			b = 16, u = 3, ++s, r = false; /* fractionless hexadecimal */
		else if ((c+='x'-'h')==0)
			b = 16, u = 3, ++s; /* hexadecimal with fraction */
		else if ((c+='h'-'o')==0) /* capital O and octal numbers with only one zero not supported */
			b = 8, ++s; /* octal; added 2007.12.08 */
		else if ((c+='o'-'b')==0)
			b = 2, ++s; /* binary; added 2007.12.09 */
		else if ((c+='b'-'d')==0)
			b = 10, r = false, ++s; /* fractionless decimal; added 2007.12.09 */
		else if ((c+='d'-'z')==0)
			b = 36, u = 3, ++s; /* base 36 (all alphanumeric); added 2007.12.09 */
		if (s>=e)
			return NULL; /* must follow alphanumeric digits */
		c = *s++;
	}
	goto start;
	do
	{
		c = *s++;
	start:
		if ((c+=-'0')<=(_uint8)('9'-'0')) /* 0-9 */
		{
			if (c>=(_uint8)b)
				return NULL; /* digit non-existent in this base number */
			u |= 0x8000;
			if (a>0)
				x = 10*x+c;
			else
				v = b*v+c, m += a;
		} else if ((c+='0'-'A')<='Z'-'A' || (c+='A'-'a')<='z'-'a' && (_uint8)u>2)/* && a<=0)*/
		{
			if ((c+=10)>=(_uint8)b)
				return NULL; /* alphanumeric digit non-existent in this base number */
			v = b*v+c, m += a, u |= 0x8000;
			continue;
		} else if ((c+='A'-'E')==0) /* Ee */
		{ /* can't have more than one `E' */
			if (a==1)
				return NULL;
			a = 1;
		} else if ((c=s[-1]-'.')==0) /* . */
		{ /* can't have more than one point or during integer parsing (includes notation) */
			if (a!=0 || !r)
				return NULL;
			a = -1;
#if MINUS_CHAR!='-'
		} else if ((c+='.'-'-')==0 || c==(_uint8)(MINUS_CHAR-'-'))
#else
		} else if ((c+='.'-'-')==0)
#endif
		{ /* only one '-' valid (here) for notation; cannot come after `E+' */
			if (a<=0 || (u&1)!=0)
				return NULL;
			u = 1;
		} else if ((c+='-'-'+')==0) /* + */
		{ /* only one '+' valid (here) for notation; cannot come after `E-' */
			if (a<=0 || (_uint8)u!=0)
				return NULL;
			u = 2;
		} else if ((c+='+'-'_')==0)
			return NULL; /* can't directly follow an underscore */
		else /* everything else ends */
			break;
	} while(s<e);

	if ((u&0x8000)==0)
		return NULL; /* no digit chars after start or notation ("...E+" | "...E-") */
	if ((_uint8)u==1)
		m -= x;
	else
		m += x;
	if (m!=0)
		v *= pow(b,m);

	if (_v!=NULL)
		*_v = n==0 ? v:-v;

	return s;/*s>e ? e:s;*/
}
#endif/*defined(env68k) && !defined(envfpu)*/
